require "cgi"
require "rd/rdvisitor"
require "rd/version"
require "rw-lib"
require "rd2rwiki-ext"

module RD
  class Reference
    class RWikiLabel < Label
      attr_reader :wikiname
      alias element_label wikiname
      alias to_label wikiname
      
      def initialize(wikiname)
	@wikiname = wikiname
      end

      def RWikiLabel.new_from_RDLabel(rdlabel)
	return RWikilabel.new(nil) if
	  rdlabel.element_label.nil? && rdlabel.filename.nil?

	prefix = rdlabel.filename ? rdlabel.filename + '/' : ""
	return RWikiLabel.new(prefix + rdlabel.element_label)
      end
      
      def result_of_apply_method_of(visitor, reference, children)
	visitor.apply_to_Reference_with_RWikiLabel(reference, children)
      end

      def self.replace_label(reference)
       	if reference.label.is_a? Reference::RDLabel
  	  new_label = RWikiLabel.new_from_RDLabel(reference.label)
   	else
  	  new_label = reference.label
   	end

    	reference.label = new_label
      end
    end
  end
end


module RD
  class RD2RWikiVisitor < RDVisitor
    include MethodParse

    SYSTEM_NAME = "RDtool -- RD2RWikiVisitor"
    SYSTEM_VERSION = "Based on RD2HTMLVisitor $Version: 0.6.11$" #"
    VERSION = Version.new_from_version_string(SYSTEM_NAME, SYSTEM_VERSION)

    def self.version
      VERSION
    end

    # must-have constants
    OUTPUT_SUFFIX = "html"
    INCLUDE_SUFFIX = ["html"]
    
    METACHAR = { "<" => "&lt;", ">" => "&gt;", "&" => "&amp;" }
    
    attr(:css, true)
    attr(:charset, true)
    alias charcode charset
    alias charcode= charset=
    attr(:lang, true)
    attr(:title, true)
    attr(:html_link_rel, nil)
    attr(:html_link_rev, nil)
    attr(:use_old_anchor, true)
    # output external Label file.
    attr(:output_rbl, true)

    attr_reader :footnotes
    attr_reader :foottexts

    attr_reader(:links)

    def initialize
      @css = nil
      @charset = nil
      @lang = nil
      @title = nil
      @html_link_rel = {}
      @html_link_rev = {}

      @footnotes = []
      @index = {}

      #      @use_old_anchor = nil
      @use_old_anchor = true # MUST -> nil
      @output_rbl = nil

      # For RWiki
      @links = []
      @ref_extension = []
      install_ref_extension

      super
    end

    def visit(tree)
      prepare_labels(tree, "label:")
      tree.find_all{|i| i.is_a? Reference }.each do |i|
	Reference::RWikiLabel.replace_label(i)
      end
      prepare_footnotes(tree)
      tmp = super(tree)
      # make_rbl_file(@filename) if @output_rbl and @filename
      tmp
    end

    # Creates content for RWiki, not a full-HTML instance.
    def apply_to_DocumentElement(element, content)
      content = content.join("\n")
      foottext = make_foottext
      title = document_title	# Not used for now.

      %Q[#{content}#{foottext}\n]
    end

    def document_title
      return @title if @title
      return @filename if @filename
      return ARGF.filename if ARGF.filename != "-"
      "Untitled"
    end
    private :document_title
    
    def apply_to_Headline(element, title)
      anchor = get_anchor(element)
      label = hyphen_escape(element.label)
      %Q[<h#{element.level}><a name="#{anchor}" id="#{anchor}">#{title}</a>] +
      %Q[</h#{element.level}><!-- RDLabel: "#{label}" -->]
    end

    # RDVisitor#apply_to_Include 

    def apply_to_TextBlock(element, content)
      content = content.join("")
      if (is_this_textblock_only_one_block_of_parent_listitem?(element) or
	  is_this_textblock_only_one_block_other_than_sublists_in_parent_listitem?(element))
	content.chomp
      else
	%Q[<p>#{content.chomp}</p>]	# <br/>#{content}\n ?
      end
    end

    def is_this_textblock_only_one_block_of_parent_listitem?(element)
      (element.parent.is_a?(ItemListItem) or
       element.parent.is_a?(EnumListItem)) and
	consist_of_one_textblock?(element.parent)
    end

    def is_this_textblock_only_one_block_other_than_sublists_in_parent_listitem?(element)
      (element.parent.is_a?(ItemListItem) or
       element.parent.is_a?(EnumListItem)) and
	consist_of_one_textblock_and_sublists(element.parent)
    end

    def consist_of_one_textblock_and_sublists(element)
      i = 0
      element.each_child do |child|
	if i == 0
	  return false unless child.is_a?(TextBlock)
	else
	  return false unless child.is_a?(List)
	end
	i += 1
      end
      return true
    end

    def apply_to_Verbatim(element)
      content = []
      element.each_line do |i|
	content.push(apply_to_String(i))
      end
      content_str = content.join("").chomp
      if /\A#\s*([^\n]+)\s*\n(.*)\z/m =~ content_str
	apply_VerbatimBlock_extension($1, $2) ||
	  %Q[<pre>\n#{content_str}</pre>\n]
      else
      	%Q[<pre>\n#{content_str}</pre>\n]
      end
    end
  
    def apply_to_ItemList(element, items)
      %Q[<ul>\n#{items.join("\n").chomp}\n</ul>]
    end
  
    def apply_to_EnumList(element, items)
      %Q[<ol>\n#{items.join("\n").chomp}\n</ol>]
    end

    def apply_to_DescList(element, items)
      %Q[<dl>\n#{items.join("\n").chomp}\n</dl>]
    end

    def apply_to_MethodList(element, items)
      %Q[<dl>\n#{items.join("\n").chomp}\n</dl>]
    end
    
    def apply_to_ItemListItem(element, content)
      %Q[<li>#{content.join("\n").chomp}</li>]
    end
    
    def apply_to_EnumListItem(element, content)
      %Q[<li>#{content.join("\n").chomp}</li>]
    end
    
    def consist_of_one_textblock?(listitem)
      listitem.content.size == 1 and listitem.content[0].is_a?(TextBlock)
    end
    private :consist_of_one_textblock?
    
    def apply_to_DescListItem(element, term, description)
      anchor = get_anchor(element.term)
      label = hyphen_escape(element.label)
      if description.empty?
	%Q[<dt><a name="#{anchor}" id="#{anchor}">#{term}</a></dt>] +
	%Q[<!-- RDLabel: "#{label}" -->]
      else
	%Q[<dt><a name="#{anchor}" id="#{anchor}">#{term}</a></dt>] +
	%Q[<!-- RDLabel: "#{label}" -->\n] +
	%Q[<dd>\n#{description.join("\n").chomp}\n</dd>]
      end
    end

    def apply_to_MethodListItem(element, term, description)
      term = parse_method(term)  # maybe: term -> element.term
      anchor = get_anchor(element.term)
      label = hyphen_escape(element.label)
      if description.empty?
	%Q[<dt><a name="#{anchor}" id="#{anchor}"><code>#{term}] +
	%Q[</code></a><!-- RDLabel: "#{label}" -->]
      else
	%Q[<dt><a name="#{anchor}" id="#{anchor}"><code>#{term}] +
	%Q[</code></a></dt><!-- RDLabel: "#{label}" -->\n] +
	%Q[<dd>\n#{description.join("\n")}</dd>]
      end
    end

    def apply_to_StringElement(element)
      apply_to_String(element.content)
    end
    
    def apply_to_Emphasis(element, content)
      %Q[<em>#{content.join("")}</em>]
    end
  
    def apply_to_Code(element, content)
      %Q[<code>#{content.join("")}</code>]
    end
  
    def apply_to_Var(element, content)
      %Q[<var>#{content.join("")}</var>]
    end
  
    def apply_to_Keyboard(element, content)
      %Q[<kbd>#{content.join("")}</kbd>]
    end
  
    def apply_to_Index(element, content)
      tmp = []
      element.each do |i|
	tmp.push(i) if i.is_a?(String)
      end
      key = meta_char_escape(tmp.join(""))
      if @index.has_key?(key)
	# warning?
	%Q[<!-- Index, but conflict -->#{content.join("")}<!-- Index end -->]
      else
	num = @index[key] = @index.size
	%Q[<a name="index:#{num}" id="index:#{num}">#{content.join("")}</a>]
      end
    end
    
    def apply_to_Reference_with_RWikiLabel(element, content)
      content = content.join("")
      apply_ref_extension(element_label(element), content)
    end
  
    def apply_to_Reference_with_URL(element, content)
      %Q[<a href="<%= ref_url(#{element.label.url.to_s.dump}) %>">#{content.join("")}</a>]
    end
    
    def apply_to_Footnote(element, content)
      num = get_footnote_num(element)
      raise ArgumentError, "[BUG?] #{element} is not registered." unless num
      
      add_foottext(num, content)
      %Q|<a name="footmark:#{num}" id="footmark:#{num}" | +
      %Q|href="#foottext:#{num}"><sup><small>| +
      %Q|*#{num}</small></sup></a>|
    end

    def get_footnote_num(fn)
      raise ArgumentError, "#{fn} must be Footnote." unless fn.is_a? Footnote
      i = @footnotes.index(fn)
      if i
	i + 1
      else
	nil
      end
    end

    def prepare_footnotes(tree)
      @footnotes = tree.find_all{|i| i.is_a? Footnote }
      @foottexts = []
    end
    private :prepare_footnotes

    def apply_to_Foottext(element, content)
      num = get_footnote_num(element)
      raise ArgumentError, "[BUG] #{element} isn't registered." unless num
      %|<a name="foottext:#{num}" id="foottext:#{num}" href="#footmark:#{num}">|+
	%|<sup><small>*#{num}</small></sup></a>| +
	%|<small>#{content}</small><br />|
    end

    def add_foottext(num, foottext)
      raise ArgumentError, "[BUG] footnote ##{num} isn't here." unless
	footnotes[num - 1]
      @foottexts[num - 1] = foottext
    end
    
    def apply_to_Verb(element)
      content = apply_to_String(element.content)
      %Q[#{content}]
    end

    def sp2nbsp(str)
      str.gsub(/\s/, "&nbsp;")
    end
    private :sp2nbsp
    
    def apply_to_String(element)
      meta_char_escape(element)
    end
    
    def parse_method(method)
      klass, kind, method, args = MethodParse.analize_method(method)
      
      if kind == :function
	klass = kind = nil
      else
	kind = MethodParse.kind2str(kind)
      end
      
      args.gsub!(/&?\w+;?/){ |m|
	if /&\w+;/ =~ m then m else '<var>'+m+'</var>' end }

      case method
      when "self"
	klass, kind, method, args = MethodParse.analize_method(args)
	"#{klass}#{kind}<var>self</var> #{method}#{args}"
      when "[]"
	args.strip!
	args.sub!(/^\((.*)\)$/, '\\1')
	"#{klass}#{kind}[#{args}]"
      when "[]="
	args.tr!(' ', '')
	args.sub!(/^\((.*)\)$/, '\\1')
	ary = args.split(/,/)

	case ary.length
	when 1
	  val = '<var>val</var>'
	when 2
	  args, val = *ary
	when 3
	  args, val = ary[0, 2].join(', '), ary[2]
	end

	"#{klass}#{kind}[#{args}] = #{val}"
      else
	"#{klass}#{kind}#{method}#{args}"
      end
    end
    private :parse_method

    def meta_char_escape(str)
      str.gsub(/[<>&]/) {
	METACHAR[$&]
      }
    end
    private :meta_char_escape
    
    def hyphen_escape(str)
      str.gsub(/--/, "&shy;&shy;")
    end

    def make_foottext
      return nil if foottexts.empty?
      content = []
      foottexts.each_with_index do |ft, num|
	content.push(apply_to_Foottext(footnotes[num], ft))
      end
      %|<hr />\n<p class="foottext">\n#{content.join("\n")}\n</p>|
    end
    private :make_foottext

    # RWiki Part
    private
    def install_ref_extension
      @ref_extension = []
      (methods + private_methods).sort.each do |m|
	if /^ref_ext/ =~ m
	  @ref_extension.push(m.intern)
	end
      end
      @ref_extension.push(:default_ref_ext)
    end

    def apply_VerbatimBlock_extension(label, content)
      method_name = "ext_VerbatimBlock_" + label.to_s
      if respond_to?(method_name, true)
	__send__(method_name, content)
      else
	nil
      end
    end

    def apply_ref_extension(label, content)
      @ref_extension.each do |entry|
	result = __send__(entry, label, content)
	return result if result
      end
    end

    def element_label(element)
      case element
      when RDElement
	element.to_label
      else
	element
      end
    end

    def rwiki_refer(label)
      @links.push label unless @links.include? label
      "<%= ref_name(#{label.to_s.dump}) %>" 
    end

    def rwiki_mod(label)
      meta_char_escape(label) + " (<%= modified(Page[#{label.dump}].modified)%>)"
    end

    def default_ref_ext(label, content)
      ref = rwiki_refer(label)
      mod = rwiki_mod(label)
      %Q[<a href="#{ref}" title="#{mod}">#{content}</a>]
    end

    def url_ref_ext(url, content)
      %Q[<a href="<%= ref_url(#{url.to_s.dump}) %>">#{content}</a>]
    end
  end # RD2RWikiVisitor
end # RD

$Visitor_Class = RD::RD2RWikiVisitor
